/**************************************************************
 *
 *
 * $Header: /homes/neuron/b/revow/src/mars/RCS/marsMain.c,v 1.2 1996/05/01 14:58:06 revow Exp revow $
 *
 * $Log: marsMain.c,v $
 * Revision 1.2  1996/05/01  14:58:06  revow
 * BootStrap Sample taken by actually copying samples from the
 * original data. Previously it was done by just using the wight matrix
 *
 * Revision 1.1  1996/05/01  14:04:50  revow
 * Initial revision
 *
 *
 * MODULE DESCRIPTION
 *
 *   Top level for Friedmans mars 3.5 code. Allows some command line options
 *   and also checks the training file for attribute dimension and
 *   number of training examples.
 *
 * PROJECT:	DELVE
 *
 * $Author: revow $
 *
 *************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "marsMain.h"

static char *optstring = "ht:o:l:k:n:s:";

#define USAGE_STRING    \
"Usage %s : -[options] file - Perform bagged mars using J. Friedman's MARS 3.5\n\n\nh - print this help\nt - Test data file name\no - Outfile name\nl - Interactive level (max num of variables per basis function)\n  l = 1 Additive model > 1 up to l variable interactions allowed (default 8)\nk - max number of basis functions (default 15)\nn - Number of bootstrap samples to take in the bag\n  n=0 use ll training data (no sampling)\ns - Specify a seed for the pseudo random number generator (defaults to the pid)\n"

int main(int argc, char *argv[])
{
    int		i;
    char	*tmpTrain = NULL;
    char	*tmpTest = NULL;
    char	*tmpOut = NULL;
    char        *trainFile = "train.0"; /* Train file */
    char	*testFile = "test.0";
    char	*outFile = "guess.0";
    int		mi = 8;		/* Max variables per basis function */
    int		nk = 15; 	/* Max number of basis functions */
    int         c;              /* Option character */
    extern char *optarg;        /* Next option argumnet */
    extern int  optind;         /* Index into next argument */
    FILE	*fp;
    int		nDimTrain, nDimTest;
    int		nTrain, nTest;
    int		nBoot = 50;
    mrReal	*f, *fav;	/* Predictor and averaged predictors */
    mrReal	*xTrain, *y, *xTest; /* Data and work arrays */
    mrReal	*xSample = NULL, *ySample = NULL; /* Bootstrap sample */
    mrReal	*w; 		 /* Weights for observations */
    int		*lx;		/* Restrictions on variables See mar3.5.f */
    int		rep;

    srand48(getpid());

    while ( (c = getopt(argc, argv, optstring)) != -1)
    {
      switch (c)
      {
       case 'h':
	fprintf(stderr, USAGE_STRING, argv[0]);
	exit (0);
	break;

       case 't':
	testFile = optarg;
	break;

       case 'o':
	outFile = optarg;
	break;

       case 'k':
	nk = atoi(optarg);
	break;

       case 'l':
	mi = atoi(optarg);
	break;

       case 'n':
	nBoot = atoi(optarg);
	break;

       case 's':
        srand48(atoi(optarg));
        break;

       default:
	break;
      }
    }

    if ( optind < argc)
    {
      trainFile = argv[optind];
    }

    if (NULL == (fp = fopen(trainFile, "r")))
    {
      fprintf(stderr, "Cannot open train file %s\n", trainFile);
      exit (1);
    }
      
    countFileColsAndRows(fp, &nDimTrain, &nTrain);
    readData(fp, (nDimTrain-1), nTrain, &xTrain, &y);
    fclose(fp);

    if (NULL == (fp = fopen(testFile, "r")))
    {
      fprintf(stderr, "Cannot open test file %s\n", testFile);
      exit (1);
    }

    countFileColsAndRows(fp, &nDimTest, &nTest);
    readData(fp, (nDimTest), nTest, &xTest, NULL);
    fclose(fp);

    f = (mrReal *)malloc(sizeof(*f) * nTest);
    fav = (mrReal *)malloc(sizeof(*f) * nTest);

    for (i = 0 ; i < nTest ; ++i)
    {
      fav[i] = 0.0;
    }

    nDimTrain--;

    if (nDimTrain != nDimTest)
    {
      printf("%s: The input dimensions in DELVE train file %s != dimension in test file %s\n", argv[0], trainFile, testFile);
      exit (1);
    }

    w = (mrReal *)malloc(sizeof(*w) * nTrain);
    lx = (int *)malloc(sizeof(*w) * nTrain);

    rep = 0;
    for (i = 0 ; i < nTrain ; ++i)
    {
      w[i] = 1.0;
    }

    do
    {
      /* Create a Bootstrap sample */
      if (nBoot > 0)
      {
	bootStrapSample(nDimTrain, nTrain, xTrain, y, &xSample, &ySample);
      }
      else
      {
	xSample = xTrain;
	ySample = y;
      }

      mainmars_(&nDimTrain, &nTrain, &nTest, &mi, &nk, xSample, ySample, xTest, f, w, lx);

      for (i = 0 ; i < nTest ; ++i)
      {
        fav[i] += f[i];
      }

      ++rep;
    } while (rep < nBoot);

    if (NULL == (fp = fopen(outFile, "w")))
    {
      fprintf(stderr, "Cannot open train file %s\n", trainFile);
    }
    else
    {
      if (rep > 0)
      {
        for (i = 0 ; i < nTest ; ++i)
        {
          fprintf(fp, "%g\n", fav[i] / (mrReal)rep);
        }
      }
      fclose(fp);
    }

    /* Release all arrays */
    free((void *)xTrain);
    free((void *)y);
    free((void *)xTest);
    free((void *)f);
    free((void *)fav);
    free((void *)w);
    free((void *)lx);

    exit (0);
}

/**************************************************************
 *
 *
 * NAME: countFileColsAndRows
 *
 * DESCRIPTION:
 *
 *  Counts number of columns in a file and number of
 *  lines
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
int countFileColsAndRows(
FILE		*fp,		/* in: Data file */
int		*ncol,		/* out: # of cols */
int		*nlines		/* out Estimated # lines */
)
{
    char	line[BUFSIZ];
    char	*lp, *lp1;
    long	nc = 0;		/* Column count on current line */


    /*
     * Count the number of lines in the file
     */
    *nlines = 0;
    while (fgets(line, BUFSIZ, fp))
    {
      nc = countCols(line, BUFSIZ);

      if (*nlines > 0)
      {
	if (nc != *ncol)
	{
	  fprintf(stderr, "Inconsistant number of columns on line %d\n", *nlines);
	  return(-1);
	}
      }   
      *ncol = nc;
      ++(*nlines);
    }

    rewind(fp);

}
/**************************************************************
 *
 *
 * NAME: countCols
 *
 * DESCRIPTION:
 *
 *  Counts columns of data in a string
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 * Number of columns 
 *
 *************************************************************/
static int countCols(
char		*line,		/* in: Line with data */
int		maxLen		/* in: Maximum line length */
)
{
    char	*lp, *lp1;
    int		ncol;

    lp1 =  line;
    lp = NULL;

    for (ncol = -1 ; lp != lp1 && (lp1 - line < maxLen) ; ++ncol)
    {
	lp = lp1;
	(void)strtod(lp, &lp1);
    }

    return (ncol);
}

void readData(
FILE		*fp,		/* in: Data file */
int		ncol,		/* out: # of cols */
int		nlines,	/* out Estimated # lines */
mrReal		**x,		/* i/o: Inputs */
mrReal		**y		/* i/o: Targets */
)
{
  int		i, j;

  rewind(fp);

  if (x != NULL)
  {
    *x = (mrReal *)malloc(ncol * nlines * sizeof(**x));
  }

  if (y != NULL)
  {
    *y = (mrReal *)malloc(nlines * sizeof(**y));
  }

  rewind(fp);

  for (i = 0 ; i < nlines ; ++i)
  {
    for (j = 0 ; j < ncol ; ++j)
    {
      fscanf(fp, "%g", &(*x)[j*nlines + i]);
    }

    if (y)
    {
      fscanf(fp, "%g", &(*y)[i]);
    }
  }

}

void bootStrapSample(
int		ncol,		/* in # of cols */
int		nlines,		/* in:# rows */
mrReal		*Xdata,		/* in: Data - inputs */
mrReal		*Ydata,		/* in: Data - Targets */
mrReal		**xSample,	/* out: Bootstarp sample - inputs */
mrReal		**ySample	/* out: Bootstarp sample - outputs */
)
{
  int		i, j, k;

  if (xSample != NULL)
  {
    *xSample = (mrReal *)malloc(ncol * nlines * sizeof(**xSample));
  }

  if (ySample != NULL)
  {
    *ySample = (mrReal *)malloc(nlines * sizeof(**ySample));
  }

  for (i = 0 ; i < nlines ; ++i)
  {
    k = REAL_TO_INT(drand48()*nlines-1);

    for (j = 0 ; j < ncol ; ++j)
    {
      (*xSample)[j*nlines + i] = Xdata[j*nlines + k];
    }

    (*ySample)[i] = Ydata[k];
  }

}

